--[[
    Polling for interrupts
]]

local addon_name, ns = ...

local Printer = LibStub:GetLibrary("OakPrint-1.0")(addon_name)

local classTable = {}
local nameTable = {}
local glyphTable = {}

local tContains, tinsert, CheckInteractDistance = tContains, tinsert, CheckInteractDistance
--------------------------------------------------------------------------------
-- Dataobject
--------------------------------------------------------------------------------

local ldb = LibStub:GetLibrary("LibDataBroker-1.1")
local dataobj = ldb:NewDataObject("OakInterrupts",
                        { type = "OakInterrupt" })

--------------------------------------------------------------------------------
-- Helper functions
--------------------------------------------------------------------------------

local mt = {
    __index = function(table, key)
        if key == "cooldown" then return 15
        elseif key == "glyph_cd" then return table["cooldown"]
        elseif key == "glyph_success_cd" then return table["glyph_cd"]
        else return nil
        end
    end
}

local spellTable = {
    ["Pummel"] = setmetatable({}, mt),
    ["Rebuke"] = setmetatable({}, mt),
    ["Kick"] = setmetatable({ cooldown = 15, glyph_cd = 19, glyph_success_cd = 13 }, mt),
    ["Mind Freeze"] = setmetatable({ glyph_cd = 14 }, mt),
    ["Wind Shear"] = setmetatable({ cooldown = 12, glyph_cd = 15 }, mt),
    ["Spear Hand Strike"] = setmetatable({}, mt),
    ["Skull Bash"] = setmetatable({}, mt)
}

local spellOfClass = {
    ["WARRIOR"] = spellTable["Pummel"],
    ["PALADIN"] = spellTable["Rebuke"],
    ["ROGUE"] = spellTable["Kick"],
    ["DEATHKNIGHT"] = spellTable["Mind Freeze"],
    ["SHAMAN"] = spellTable["Wind Shear"],
    ["MONK"] = spellTable["Spear Hand Strike"],
    ["DRUID"] = spellTable["Skull Bash"]
}

local GlyphAltersCooldown = {
    [56805] = true,
    [58686] = true,
    [55451] = true
}

--[[
local specs = {
    -- Mage --
    [62] = "Arcane",
    [63] = "Fire",
    [64] = "Frost",
    -- Paladin --
    [65] = "Holy",
    [66] = "Protection",
    [70] = "Retribution",
    -- Warrior --
    [71] = "Arms",
    [72] = "Fury",
    [73] = "Protection",
    -- Druid --
    [102] = "Balance",
    [103] = "Feral",
    [104] = "Guardian",
    [105] = "Restoration",
    -- Death Knight --
    [250] = "Blood",
    [251] = "Frost",
    [252] = "Unholy",
    -- Hunter --
    [253] = "Beast Mastery",
    [254] = "Marksmanship",
    [255] = "Survival",
    -- Priest --
    [256] = "Discipline",
    [257] = "Holy",
    [258] = "Shadow",
    -- Rogue --
    [259] = "Assassination",
    [260] = "Combat",
    [261] = "Subtlety",
    -- Shaman --
    [262] = "Elemental",
    [263] = "Enhancement",
    [264] = "Restoration",
    -- Warlock --
    [265] = "Affliction",
    [266] = "Demonology",
    [267] = "Destruction",
    -- Monk --
    [268] = "Brewmaster",
    [269] = "Windwalker",
    [270] = "Mistweaver"
}

]]

local ClassIsRanged = {
    MAGE = true,
    HUNTER = true,
    PRIEST = true,
    WARLOCK = true
}

local SpecIsMelee = {
    [66] = true,
    [70] = true,
    [71] = true,
    [72] = true,
    [73] = true,
    [103] = true,
    [104] = true,
    [250] = true,
    [251] = true,
    [252] = true,
    [259] = true,
    [260] = true,
    [261] = true,
    [263] = true,
    [268] = true,
    [269] = true
}

local function tremovebyval(t, val)
    for k, v in ipairs(t) do
        if v == val then
            tremove(t, k)
            break
        end
    end
end

--------------------------------------------------------------------------------
-- Metatable magic
--------------------------------------------------------------------------------

local _interruptTable = {}
local interruptTable = {}

setmetatable(interruptTable, {
    __index = _interruptTable,
    __newindex = function(t, k, v)
        assert(not (k == nil), "interruptTable: tried to set nil key")
        local name = nameTable[k]
        if v == nil then -- unregister
            _interruptTable[k] = nil
            nameTable[k] = nil
            dataobj[name] = nil
            return
        end
        local class = classTable[k]
        local spell = spellOfClass[class]
        if spell then
            local cd = (glyphTable[k] and spell.glyph_cd)
                        or spell.cooldown
            local time = v > 0 and v + cd or _interruptTable[k] or 0
            Printer.debug(2,
                       "Setting interrupt for %s with expiration time %d",
                       name, time)
            _interruptTable[k] = time
            dataobj[name] = time -- set exported data
        end
    end,
    __call = function()
        return pairs(_interruptTable)
    end
})

--------------------------------------------------------------------------------

local controller = CreateFrame("Frame")

controller:RegisterEvent("INSPECT_READY")
controller:RegisterEvent("UNIT_SPELLCAST_SUCCEEDED")
--controller:RegisterEvent("RAID_ROSTER_UPDATE")
controller:RegisterEvent("GROUP_ROSTER_UPDATE")
controller:RegisterEvent("PLAYER_SPECIALIZATION_CHANGED")

--------------------------------------------------------------------------------
-- Inspection
--------------------------------------------------------------------------------

local function HasRelevantGlyph(name)
    local num_glyphs = GetNumGlyphSockets()
    for index = 1, num_glyphs do
        local enabled, g_type, _, g_id = GetGlyphSocketInfo(index, nil, true, name)
        return GlyphAltersCooldown[g_id]
    end
end

local function HasRelevantSpec(name)
    Printer.debug(2, "Inspecting %s", name)
    local specID = GetInspectSpecialization(name)
    Printer.debug(2, "Got specID %d", specID)
    return SpecIsMelee[specID]
end


local notifyRegister = {} 
local oorRegister = {}
local inspecting = false
local inspect_timeout

local function RegisterInspection(name)
    Printer.debug(2, "Registering %s for inspection", name)
    if not tContains(notifyRegister, name) then
        table.insert(notifyRegister, name)
    end
    if not inspecting then
        Printer.debug(3, "Next inspection is for %s (calling NotifyInspect)", name)
        NotifyInspect(name)
        inspect_timeout = 4
        inspecting = UnitGUID(name)
    end
end

local next_oorpass = 2
local function OoRInspection(self, elapsed)
    if elapsed > next_oorpass then
        local switching = {}
        for _, v in ipairs(oorRegister) do
            if CheckInteractDistance(v, 1) then
                tinsert(switching, v)
            end
        end
        for _, v in ipairs(switching) do
            tremovebyval(oorRegister, v)
            RegisterInspection(v)
        end
        next_oorpass = 2
        --[[
        if #oorRegister == 0 then
            Printer.debug(2, "Stopping out of range polling")
            controller:SetScript("OnUpdate", nil)
        end
        ]]
    else
        next_oorpass = next_oorpass - elapsed
    end
end

local function InspectNext()
    --[[
    if #notifyRegister == 0 and #oorRegister > 0 then
        Printer.debug(2, "Starting out of range polling")
        controller:SetScript("OnUpdate", OoRInspection)
        return
    end
    ]]
    for k, name in ipairs(notifyRegister) do
        if CheckInteractDistance(name, 1) then
            RegisterInspection(name)
            break
        else
            Printer.debug(2, "%s is too far away to inspect", name)
            tinsert(oorRegister, 1, name)
        end
    end
    for k, name in ipairs(oorRegister) do
        tremovebyval(notifyRegister, name)
    end
end

local function InspectGUID(guid)
    local name = nameTable[guid]
    Printer.debug(2, "Name: %s, GUID: %s", name, guid)
    assert(name, "No name associated with GUID " .. guid)

    if not HasRelevantSpec(name) then
        Printer.debug(2, "%s has irrelevant spec", name)
        interruptTable[guid] = nil --"UNREGISTER"
    else
        if not interruptTable[guid] then
            interruptTable[guid] = 0
        end
        oorRegister[name] = nil
        glyphTable[guid] = HasRelevantGlyph(name)
        tremovebyval(notifyRegister, name)
    end
    ClearInspectPlayer()
    inspecting = false
end

--------------------------------------------------------------------------------
-- Polling
--------------------------------------------------------------------------------

local function OnUpdate(self, elapsed)
    if inspecting and inspect_timeout and inspect_timeout < elapsed then
        ClearInspectPlayer()
        inspecting = false
        InspectNext()
    end
    OoRInspection(self, elapsed)
end

local function PollByUnitID(unitID, force)
    local name = UnitName(unitID)
    local guid = UnitGUID(unitID)
    if not guid or not name or name == "Unknown" then return end
    if force or not interruptTable[guid] then
        local class = select(2, UnitClass(unitID))
        if ClassIsRanged[class] then
            Printer.debug(2, "%s is definitely ranged; not processing", name)
            return
        end
        Printer.debug(2, "Processing %s:%s", name, guid)
        nameTable[guid] = name
        classTable[guid] = class
        RegisterInspection(name)
    end
end

local function TestPoll()
    Printer.debug(1, "Running test poll")
    PollByUnitID('player', true)
end

local function PurgeUnit(guid)
    local name = nameTable[guid]
    Printer.debug(3, "Purging %s:%s", name, guid)
    interruptTable[guid] = nil
    notifyRegister[name] = nil
    oorRegister[name] = nil
end

local function PollRaid()
    Printer.debug(1, "Removing players who left raid")
    for guid, _ in interruptTable() do
        if not UnitInRaid(nameTable[guid]) then
            PurgeUnit(guid)
        end
    end
    if IsInRaid() then
        Printer.debug(1, "Polling raid group")
        local num = GetNumGroupMembers()
        for i=1,num do
            local unit = 'raid' .. i
            PollByUnitID(unit)
        end
    end
    -- finalling we need to kick off the inspect sequence
    --InspectNext()
    inspect_timeout = 1
end

local function handleEvents(self, event, ...)
        Printer.debug(5, "Event caught: %s\n    args: %s", event, strjoin(", ", ...))
        if event == "UNIT_SPELLCAST_SUCCEEDED" then
            local unitID, spell, _, lineID, spellID = ...
            local guid = UnitGUID(unitID)
            if type(interruptTable[guid]) == "number" and spellTable[spell] then
                Printer.debug(5, "Tracked spell (%s) cast by %s", spell, nameTable[guid])
                interruptTable[guid] = GetTime()
            end
        elseif event == "INSPECT_READY" then
            local guid = ...
            --if nameTable[guid] and tContains(notifyRegister, nameTable[guid]) then
            if inspecting == guid then
                InspectGUID(guid)
                InspectNext()
            else
                Printer.debug(3, "%s is not the current inspection target", UnitName(guid) or guid)
            end
        elseif event == "GROUP_ROSTER_UPDATE" then
            Printer.debug(2, "Roster Update")
            PollRaid()
        elseif event == "PLAYER_SPECIALIZATION_CHANGED" then
            local unitID = ...
            Printer.debug(2, "%s changed spec", unitID)
            if UnitInRaid(unitID) then
                PollByUnitID(unitID, true)
            end
        end
    end
controller:SetScript("OnEvent", handleEvents)
controller:SetScript("OnUpdate", OnUpdate)
--------------------------------------------------------------------------------
-- Slash commands
-------------------------------------------------------------------------------- 

local function Dump(heading, t)
    if t == dataobj then
        _, t = ldb:pairs(t)
    end
    Printer.print("--- " .. heading .. " ---")
    for k, v in next, t do
       Printer.print("%s: %s", k, tostring(v))
    end
end

local commands = LibStub:GetLibrary("OakSlash-1.0")(addon_name)
local t = {
    poll = PollRaid,
    test = TestPoll,
    dump = { all = function()
                       Dump("Exported Data", dataobj)
                       Dump("Notify Register", notifyRegister)
                       Dump("Out of Range", oorRegister)
                   end,
             data = function() Dump("Exported Data", dataobj) end,
             inspect = function()
                       Dump("Current inspection target", {UnitName(inspecting)})
                       Dump("Notify Register", notifyRegister)
                       Dump("Out of Range", oorRegister)
                   end},
    debug = Printer.SetDebug
}
commands:NewCommand("oib", t)

-- TODO:
